/*
 * bundle.c: url bundle file handling
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Mohd Husaam Mehdi <husaam.mehdi@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <ctype.h>
#include <curl/curl.h>

#include "bundle.h"

/*
 * Function to check whether a string is a proper domain/url
 * Valid domain: example.com
 * Valid domain: subdomain.example.com
 * Valid domain: my-domain.net
 * Valid domain: example-123.org
 * Invalid domain: example..com
 * Invalid domain: example@domain.com
 * Invalid domain: example
 * Invalid domain: 127.0.0.1
 */
int is_valid_domain(const char *line)
{
	if (!line || !strlen(line))
		return 0;

	enum { Start, InLabel, Dot, TLD, Error } state = Start;
	const char *ptr = line + strlen(line) - 1;  // Start from the last character
	int tld_length = 0;
	int label_length = 0;
	int num_labels = 0;

	// Traverse backward through the string
	while (ptr >= line) {
		char ch = *ptr;

		switch (state) {
		case Start:
			if (isalpha(ch)) {
				state = TLD;  // Start with TLD if it ends with a dot
				tld_length++;
			} else {
				state = Error;     // Invalid start character
			}

		case InLabel:
			if (isalnum(ch)) {
				label_length++;       // Stay in label
			} else if (ch == '-') {
				if (label_length == 0 || ptr == line) {
					state = Error;  // Hyphen cannot start a label
				} else {
					label_length++;  // Hyphen within label is okay
				}
			} else if (ch == '.') {
				// Ensure label does not end with a hyphen
				if (*(ptr - 1) == '-') {
					state = Error;
				} else {
					state = Dot;  // Move to dot state
					label_length = 0;
				}
			} else {
				state = Error;  // Invalid character in label
			}
			break;

		case Dot:
			if (isalnum(ch)) {
				num_labels++;
				state = InLabel;   // Start of a new label
				label_length = 1;
			} else if (ch == '*') {
				if (num_labels < 1) {
					state = Error;
				} else {
					state = InLabel;
				}
			} else {
				state = Error;  // Invalid character after dot
			}
			break;

		case TLD:
			if (isalpha(ch)) {
				tld_length++;  // Alphabetic characters only in TLD
			} else if (ch == '.') {
				if (tld_length < 2) {
					state = Error;     // Invalid start character
				} else {
					state = Dot;  // Expect label after `*.`
				}
			} else {
				state = Error;  // Invalid TLD structure
			}
			break;

		case Error:
			return 0;  // Invalid domain
		}
		ptr--;  // Move to the previous character
	}

	// Final check for valid TLD length and label before TLD
	return (state == InLabel && tld_length >= 2 && num_labels >= 1);
}

// Callback function used by libcurl to write data into the buffer
static size_t write_callback(void *contents, size_t size, size_t nmemb, void *userp)
{
	size_t total_size = size * nmemb;
	struct dynamic_memory *mem = (struct dynamic_memory *)userp;

	// Check if there's enough capacity to add the new data
	if (mem->size + total_size > mem->capacity) {
		return 0;
	}

	// Copy the new data to the buffer
	memcpy(&(mem->data[mem->size]), contents, total_size);
	mem->size += total_size;

	// Null-terminate the buffer
	mem->data[mem->size] = 0;

	return total_size;
}

// Function to get the size of the file from the URL using a HEAD request
size_t get_content_length(const char *url)
{
	CURL *curl_handle;
	CURLcode res;
	curl_off_t content_length = 0;

	curl_global_init(CURL_GLOBAL_DEFAULT);
	curl_handle = curl_easy_init();
	if (curl_handle) {
		// Set URL
		curl_easy_setopt(curl_handle, CURLOPT_URL, url);
		// Set a HEAD request
		curl_easy_setopt(curl_handle, CURLOPT_NOBODY, 1L);
		// Follow redirects if necessary
		curl_easy_setopt(curl_handle, CURLOPT_FOLLOWLOCATION, 1L);

		// Perform the request
		res = curl_easy_perform(curl_handle);
		if (res == CURLE_OK) {
			// Get the content length from the header
			res = curl_easy_getinfo(curl_handle, CURLINFO_CONTENT_LENGTH_DOWNLOAD_T, &content_length);
			if (res != CURLE_OK || content_length <= 0) {
				content_length = 0; // Fallback in case we don't get a valid length
			}
		} else {
			log_error2("curl_easy_perform() failed:", curl_easy_strerror(res));
		}

		// Cleanup
		curl_easy_cleanup(curl_handle);
	}

	curl_global_cleanup();

	return (size_t)content_length;
}

// Function to download a file from a URL using libcurl
int download_file(const char *url, struct dynamic_memory *mem)
{
	CURL *curl_handle;
	CURLcode res;

	mem->size = 0;  // Initialize buffer size to zero

	curl_global_init(CURL_GLOBAL_DEFAULT);
	curl_handle = curl_easy_init();
	if (curl_handle) {
		curl_easy_setopt(curl_handle, CURLOPT_URL, url);
		curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, write_callback);
		curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)mem);

		res = curl_easy_perform(curl_handle);
		if (res != CURLE_OK) {
			log_error2("curl_easy_perform() failed:", curl_easy_strerror(res));
			return 1;
		}

		curl_easy_cleanup(curl_handle);
	}

	curl_global_cleanup();

	return 0;
}

// Function to populate the download_url array with URLs from the downloaded file
void populate_url_stringstore_from_memory(struct dynamic_memory *mem, struct urlbundle *bundle)
{
	if (!mem || !bundle || !mem->data) {
		log_error("null pointers provided!");
		return;
	}

	if (!mem->size) {
		log_error("download buffer empty!");
		return;
	}

	// Ensure the data is null-terminated (it should be, but this is a safeguard)
	mem->data[mem->size - 1] = '\0';  // Safeguard by making sure the last byte is null

	// Step 1: Split mem->data into an array of lines (strings)
	unsigned long num_lines = 0;
	for (size_t i = 0; i < mem->size; i++) {
		if (mem->data[i] == '\n') num_lines++;
	}

	// Increment num_lines for the last line if it doesn't end with a newline
	num_lines++;

	// Allocate an array to hold each line as a separate string
	char **lines = (char **)calloc(num_lines, sizeof(char *));
	if (!lines) {
		log_error("Memory allocation failed");
		return;
	}

	size_t line_index = 0;
	char *line_start = mem->data;
	for (size_t i = 0; i < mem->size; i++) {
		if (mem->data[i] == '\n') {
			mem->data[i] = '\0';  // Replace newline with null terminator
			lines[line_index++] = line_start;
			line_start = &mem->data[i + 1];
		}
	}

	// Add the last line if it doesn't end with a newline
	if (line_start < &mem->data[mem->size]) {
		lines[line_index] = line_start;
	}

	// Step 2: Use cmph_io_vector_adapter to create the Hash function
	cmph_io_adapter_t *source = cmph_io_vector_adapter((char **)lines, (cmph_uint32)num_lines);
	if (!source) {
		log_error("Error creating vector adapter");
		free(lines);
		return;
	}

	cmph_config_t *config = cmph_config_new(source);
	cmph_config_set_algo(config, CMPH_BDZ);
	CMPH_HASH hash_type = CMPH_HASH_JENKINS;
	cmph_config_set_hashfuncs(config, &hash_type);

	cmph_t *hash = cmph_new(config);
	if (!hash) {
		log_error("Error creating Hash function");
		cmph_config_destroy(config);
		cmph_io_vector_adapter_destroy(source);
		free(lines);
		return;
	}

	cmph_config_destroy(config);

	stringstore_init(&bundle->download_url, num_lines, (unsigned long)mem->size);
	if (!bundle->download_url.data) {
		log_error("failed to assign memory to download_url");
		return;
	}

	unsigned long valid_count = 0;
	// Step 3: Populate the urlbundle using the Hash function
	for (unsigned long i = 0; i < num_lines; i++) {
		if (is_valid_domain(lines[i])) {
			unsigned long id = cmph_search(hash, lines[i], (cmph_uint32)strlen(lines[i]));
			if (id < num_lines) {
				valid_count++;
				stringstore_append_string(&bundle->download_url, id, lines[i], strlen(lines[i]));
			} else {
				char buf[16] = {0};
				snprintf(buf, sizeof(buf), "%lu", id);
				log_error2("Warning: Hash ID out of bounds", buf);
			}
		}
	}

	// Save the hash and configuration to the bundle
	bundle->download_url.cmph_hash = hash;

	// Cleanup
	cmph_io_vector_adapter_destroy(source);
	free(lines);

	syslog(LOG_INFO, "Added %lu urls from downloaded file\n", valid_count);
}

void process_bundle_file(const char *file_path, const char *output_file,
			 unsigned long *line_count,
			 unsigned long *byte_count)
{
	if (!file_path || !strlen(file_path) || !output_file || !strlen(output_file) || !line_count || !byte_count) {
		log_error("insufficient arguments");
		return;
	}

	syslog(LOG_INFO, "Processing file: %s\n", file_path);

	char line[BUFSIZE] = {0};
	FILE *fp = NULL;

	fp = fopen(file_path, "r");
	if (fp == NULL) {
		log_error2("File pointer empty for", file_path);
		return;
	}

	FILE *output_fp = fopen(output_file, "w");
	if (!output_fp) {
		fclose(fp);
		log_error("Failed to open output file");
		return;
	}

	unsigned long valid_count = 0;
	while (fgets(line, sizeof(line), fp)) {
		line[strcspn(line, "\n\r")] = 0;  // Remove trailing newline or carriage return
		if (is_valid_domain(line)) {
			*byte_count += strlen(line) + 1; // +1 for newline

			fputs(line, output_fp);
			fputc('\n', output_fp); // Add newline for the final line
			valid_count++;
		}
	}

	char buf[16] = {0};
	snprintf(buf, sizeof(buf), "%lu", valid_count);

	fclose(fp);
	fclose(output_fp);

	*line_count += valid_count;
}

void populate_url_stringstore_from_file(const char *filename, struct urlbundle *bundle,
					unsigned long line_count,
					unsigned long byte_count)
{
	if (!filename || !bundle || !line_count || !byte_count)
		return;

	FILE *keys_fd = fopen(filename, "r");
	if (!keys_fd) {
		log_error2("Error opening input file", filename);
		return;
	}

	cmph_io_adapter_t *source = cmph_io_nlfile_adapter(keys_fd);
	cmph_config_t *config = cmph_config_new(source);
	cmph_config_set_algo(config, CMPH_BDZ);
	CMPH_HASH hash_type = CMPH_HASH_JENKINS;
	cmph_config_set_hashfuncs(config, &hash_type);

	cmph_t *hash = cmph_new(config);
	if (!hash) {
		log_error2("Error creating Hash function for", filename);
		cmph_config_destroy(config);
		cmph_io_nlfile_adapter_destroy(source);
		fclose(keys_fd);
		return;
	}

	cmph_config_destroy(config);

	rewind(keys_fd);

	// stringstore stuff starts
	char read_buffer[MAX_URL_LENGTH] = {0};

	// Allocate the memory for the resource
	stringstore_init(&bundle->download_url, line_count, byte_count);
	if (!bundle->download_url.data) {
		log_info("failed to assign memory to download_url");
		return;
	}

	while (fgets(read_buffer, sizeof(read_buffer), keys_fd)) {
		read_buffer[strcspn(read_buffer, "\n\r")] = 0;  // Remove trailing newline or carriage return
		unsigned long id = cmph_search(hash, read_buffer, (cmph_uint32)strlen(read_buffer));
		if (id < line_count) {
			stringstore_append_string(&bundle->download_url, id, read_buffer, strlen(read_buffer));
		} else {
			char buf[16] = {0};
			memset(buf, '\0', sizeof(buf));

			snprintf(buf, sizeof(buf), "%lu", id);
			log_error2("Warning: Hash ID out of bounds", buf);
		}
	}

	cmph_io_nlfile_adapter_destroy(source);
	fclose(keys_fd);

	bundle->download_url.cmph_hash = hash;

	syslog(LOG_INFO, "Added %lu urls [%lu bytes]\n", line_count, byte_count);
}
