/*
 * bundle.c: url bundle file handling
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Mohd Husaam Mehdi <husaam.mehdi@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include "stringstore.h"
#include "config.h"

// Initialize the stringstore
void stringstore_init(struct stringstore *store, unsigned long line_count, unsigned long byte_count)
{
	// Allocate a large contiguous block of memory to hold all strings
	store->data = calloc(byte_count, sizeof(char));
	if (!store->data) {
		log_error("Failed to allocate memory for data");
		return;
	}

	// Allocate an array of pointers to each string within data
	store->string_arr = calloc(line_count, sizeof(char *));
	if (!store->string_arr) {
		log_error("Failed to allocate memory for string pointer array");
		free(store->data);
		return;
	}

	// Initialize other fields
	store->num_strings = line_count;
	store->byte_count = byte_count;
	store->offset = 0;

	store->cmph_hash = NULL;
}

// Append a string to the stringstore at the given ID
void stringstore_append_string(struct stringstore *store, unsigned long id, const char *string, size_t length)
{
	// Ensure ID is within bounds
	if (id >= store->num_strings) {
		return;
	}

	// Check if there's enough space in the data block for the string and the null terminator
	if (store->offset + length + 1 > store->byte_count) { // Use byte_count for overflow prevention
		return;
	}

	// Copy the string into the data block at the current offset
	char *destination = store->data + store->offset;
	memcpy(destination, string, length);
	destination[length] = '\0'; // Null-terminate the string

	// Store the pointer to the string at the correct index in the ptr array
	store->string_arr[id] = destination;

	// Update the offset
	store->offset += length + 1; // +1 for the null terminator
}

bool stringstore_check_string(stringstore *str_store, char *string)
{
	if (!str_store || !string || strlen(string) == 0 || !str_store->string_arr || !str_store->cmph_hash || !str_store->num_strings) {
		return false;
	}

	unsigned int id = cmph_search(str_store->cmph_hash, string, (cmph_uint32)strlen(string));

	if (id < str_store->num_strings) {
		if (str_store->string_arr[id] && !strcmp(string, str_store->string_arr[id])) {
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

void stringstore_free(stringstore *s)
{
	if (s) {
		if (s->string_arr)
			free(s->string_arr);

		if (s->data)
			free(s->data);

		s->num_strings = 0;

		if (s->cmph_hash) {
			cmph_destroy(s->cmph_hash);
			s->cmph_hash = NULL;
		}
	}
}
