/*
 * ubus.c: URL filter ubus method
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <libubus.h>
#include <syslog.h>
#include <dirent.h>

#include "filter.h"
#include "dns_cache.h"
#include "utils.h"
#include "ubus.h"

struct ubus_context *ctx = NULL;
static const char *ubus_path;

void print_url_list(struct list_head *lh, struct blob_buf *b)
{
	if (!lh)
		return;

	struct url_list *url;

	list_for_each_entry(url, lh, list) {
		void *table = blobmsg_open_table(b, "");

		blobmsg_add_string(b, "url", url->url);
		blobmsg_add_u64(b, "hits", url->match_cnt);

		blobmsg_close_table(b, table);
	}
}

static int dump_dns_cache(struct ubus_context *context, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg)
{
	void *cache_array = NULL, *cache_table = NULL;
	void *ipv4_array = NULL, *ipv4_table = NULL;
	void *ipv6_array = NULL, *ipv6_table = NULL;
	void *cname_array = NULL, *cname_table = NULL;
	struct blob_buf b = {0};
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	cache_array = blobmsg_open_array(&b, "dns_cache");
	for (size_t i = 0; i < dns_cache_cnt; i++) {
		cache_table = blobmsg_open_table(&b, "");

		dns_cache *cache = &global_dns_cache[i];
		blobmsg_add_string(&b, "url", cache->url);
		blobmsg_add_u64(&b, "hits", (uint64_t)cache->match_cnt);

		if (cache->ip_addr_cnt > 0) {
			ipv4_array = blobmsg_open_array(&b, "ipv4_addr");
			for (size_t j = 0; j < cache->ip_addr_cnt; j++) {
				ipv4_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "addr", cache->ip_addr[j]);
				blobmsg_close_table(&b, ipv4_table);
			}
			blobmsg_close_array(&b, ipv4_array);
		}

		if (cache->ipv6_addr_cnt > 0) {
			ipv6_array = blobmsg_open_array(&b, "ipv6_addr");
			for (size_t j = 0; j < cache->ipv6_addr_cnt; j++) {
				ipv6_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "addr", cache->ipv6_addr[j]);
				blobmsg_close_table(&b, ipv6_table);
			}
			blobmsg_close_array(&b, ipv6_array);
		}

		if (cache->cname_cnt > 0) {
			cname_array = blobmsg_open_array(&b, "cname");
			for (size_t j = 0; j < cache->cname_cnt; j++) {
				cname_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "name", cache->cname[j]);
				blobmsg_close_table(&b, cname_table);
			}
			blobmsg_close_array(&b, cname_array);
		}

		blobmsg_close_table(&b, cache_table);
	}

	blobmsg_close_array(&b, cache_array);

	cache_array = blobmsg_open_array(&b, "url_list");
	struct profile *p = NULL;

	// Iterate over the global profile list
	// so, we loop over all filters of all profiles
	// to cover all possible urls
	list_for_each_entry(p, &profiles, list) {
		struct urlfilter *filter = NULL;

		list_for_each_entry(filter, &p->filters, list) {
			// Check if the filter is enabled
			if (!filter->enable) {
				continue;
			}

			print_url_list(&filter->filter_text, &b);

			if (filter->bundle_cnt) {
				for (unsigned int i = 0; i < filter->bundle_cnt; i++) {
					print_url_list(&filter->bundles[i]->custom_url, &b);
				}
			}
		}
	}
	blobmsg_close_array(&b, cache_array);

	cache_array = blobmsg_open_array(&b, "dns_server");
	for (unsigned int i = 0; i < dns_server_cnt; i++) {
		cache_table = blobmsg_open_table(&b, "");
		blobmsg_add_string(&b, "server", dns_server_ip[i]);
		blobmsg_close_table(&b, cache_table);
	}
	blobmsg_close_array(&b, cache_array);

	ubus_send_reply(context, req, b.head);

	blob_buf_free(&b);
	return 0;
}

static int show_config(struct ubus_context *context, struct ubus_object *obj,
		       struct ubus_request_data *req, const char *method,
		       struct blob_attr *msg)
{
	void *profile_array = NULL;
	struct blob_buf b = {0};
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	profile_array = blobmsg_open_array(&b, "profiles");
	struct profile *prof = NULL;
	list_for_each_entry(prof, &profiles, list) {
		print_profile(prof, &b);
	}

	blobmsg_close_array(&b, profile_array);

	ubus_send_reply(context, req, b.head);

	blob_buf_free(&b);
	return 0;
}

int list_files(const char *directory_path, char *buffer, size_t buffer_length)
{
	if (!directory_path || !buffer) {
		return 0; // Invalid arguments
	}

	DIR *dir = opendir(directory_path);
	if (!dir) {
		perror("Failed to open directory");
		return 0; // Failed to open directory
	}

	struct dirent *entry;
	size_t used_length = 0;
	int file_count = 0;

	while ((entry = readdir(dir)) != NULL && file_count < 20) {
		// Skip "." and ".."
		if (strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0) {
			continue;
		}

		// Use snprintf to safely append the file name
		int written = snprintf(buffer + used_length, buffer_length - used_length,
				       "%s%s", (file_count > 0 ? "," : ""), entry->d_name);

		// Check if snprintf succeeded and didn't exceed buffer length
		if (written < 0 || (size_t)written >= buffer_length - used_length) {
			break; // Buffer is full
		}

		used_length += (unsigned int)written; // Update used buffer length
		file_count++;
	}

	closedir(dir);
	return file_count; // Return the number of files listed
}

static int default_bundles(struct ubus_context *context, struct ubus_object *obj,
			   struct ubus_request_data *req, const char *method,
			   struct blob_attr *msg)
{
	struct blob_buf b = {0};
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	const char *path = "/tmp/urlfilter/default/urlbundles/";
	char files_buf[BUFSIZE] = {0};

	if (list_files(path, files_buf, sizeof(files_buf)))
		blobmsg_add_string(&b, "default_bundles", files_buf);
	else
		blobmsg_add_string(&b, "default_bundles", "N/A");

	ubus_send_reply(context, req, b.head);

	blob_buf_free(&b);
	return 0;
}

static struct ubus_method urlfilter_methods[] = {
	UBUS_METHOD_NOARG("dump", dump_dns_cache),
	UBUS_METHOD_NOARG("show_config", show_config),
	UBUS_METHOD_NOARG("default_bundles", default_bundles),
};

static struct ubus_object_type urlfilter_type =
	UBUS_OBJECT_TYPE("urlfilter", urlfilter_methods);

static struct ubus_object urlfilter_object = {
	.name = "urlfilter",
	.type = &urlfilter_type,
	.methods = urlfilter_methods,
	.n_methods = ARRAY_SIZE(urlfilter_methods),
};

static int urlfilter_publish_object(struct ubus_context *context)
{
	int ret;

	ret = ubus_add_object(context, &urlfilter_object);
	if (ret)
		fprintf(stderr, "Failed to add 'urlfilter' ubus object: %s\n",
			ubus_strerror(ret));

	return ret;
}

int urlfilter_ubus_init(const char *path)
{
	uloop_init();
	ubus_path = path;

	ctx = ubus_connect(path);
	if (!ctx) {
		log_error("Failed to connect to ubus");
		return -EIO;
	}

	ubus_add_uloop(ctx);

	if (urlfilter_publish_object(ctx)) {
		log_error("Failed to publish ubus object");
		return -1;
	}

	return 0;
}

void urlfilter_ubus_done(void)
{
	ubus_free(ctx);
}

